# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library to SSH into a Cloud Run Deployment."""

import argparse
from collections.abc import Sequence
import dataclasses
import enum
import json
import subprocess
from typing import Any

from googlecloudsdk.api_lib.run import constants
from googlecloudsdk.api_lib.run import k8s_object
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import iap_tunnel
from googlecloudsdk.command_lib.compute import ssh_utils
from googlecloudsdk.command_lib.util.ssh import ssh
from googlecloudsdk.core import execution_utils
from googlecloudsdk.core import log
from googlecloudsdk.core import requests as core_requests
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.util import platforms


def _GetProjectNumberFromWorkloadJson(
    workload_json: dict[str, Any],
) -> str:
  """Retrieves the project number from the workload JSON.

  Args:
    workload_json: dict, The JSON representation of the Cloud Run workload.

  Returns:
    str, The project number.

  Raises:
    ValueError: If the project number is not found in the workload JSON.
  """
  project_number = workload_json.get(constants.METADATA, {}).get(
      constants.NAMESPACE
  )
  if not project_number:
    raise ValueError("Project number not found in workload.")
  return project_number


def _ValidateGen2(workload_json: dict[str, Any]) -> None:
  """Validates that the workload is gen2.

  Args:
    workload_json: dict, The JSON representation of the Cloud Run workload.

  Raises:
    ValueError: If the workload is a gen1 deployment.
  """
  template = workload_json.get(constants.SPEC, {}).get(constants.TEMPLATE, {})
  execution_environment = (
      template.get(constants.METADATA, {})
      .get(constants.ANNOTATIONS, {})
      .get(k8s_object.EXECUTION_ENVIRONMENT_ANNOTATION)
  )
  if execution_environment == constants.GEN1:
    raise ValueError(
        "SSH is not supported for Cloud Run gen1 deployments. If you"
        " already switched the execution environment to gen2, please wait a"
        " few minutes for the deployment to be updated."
    )


def _ValidateSSHEnabled(workload_json: dict[str, Any]) -> None:
  """Validates that the workload has SSH enabled.

  Args:
    workload_json: dict, The JSON representation of the Cloud Run workload.

  Raises:
    ValueError: If the workload has SSH disabled.
  """
  ssh_enabled = (
      workload_json.get(constants.METADATA, {})
      .get(constants.ANNOTATIONS, {})
      .get(k8s_object.SSH_ENABLED_ANNOTATION)
  )
  if ssh_enabled != "true":
    raise ValueError(
        "SSH is not enabled for this deployment. Use the --ssh flag to enable"
        " SSH for this deployment."
    )


def CreateSshTunnelArgs(
    track,
    project_number,
    project,
    deployment_name,
    workload_type,
    region,
    instance_id=None,
    container_id=None,
    iap_tunnel_url_override=None,
):
  """Construct an SshTunnelArgs from command line args and values.

  Args:
    track: ReleaseTrack, The currently running release track.
    project_number: str, the project number (string with digits).
    project: str, the project id.
    deployment_name: str, the name of the deployment. For services this will be
        of the form {service-name}/revisions/{revision-id}.
    workload_type: Ssh.WorkloadType, the type of the workload.
    region: str, the region of the deployment.
    instance_id: str, the instance id (optional).
    container_id: str, the container id (optional).
    iap_tunnel_url_override: str, the IAP tunnel URL override (optional).

  Returns:
    SshTunnelArgs.
  """

  cloud_run_args = {}
  cloud_run_args["deployment_name"] = deployment_name
  cloud_run_args["workload_type"] = workload_type
  cloud_run_args["project_number"] = project_number
  cloud_run_args["instance_id"] = instance_id
  cloud_run_args["container_id"] = container_id

  res = iap_tunnel.SshTunnelArgs()
  res.track = track.prefix
  res.cloud_run_args = cloud_run_args
  res.region = region
  res.project = project

  if iap_tunnel_url_override:
    res.pass_through_args.append(
        f"--iap-tunnel-url-override={iap_tunnel_url_override}"
    )

  return res


# Double quotes around the type hint are used to avoid circular dependency
@dataclasses.dataclass(frozen=True)
class SshCommandComponents:
  """The components of an SSH command."""

  env: "ssh.Environment"
  remote: "ssh.Remote"
  cert_file: str
  iap_tunnel_args: "iap_tunnel.SshTunnelArgs"
  options: dict[str, str]
  identity_file: str


class Ssh:
  """SSH into a Cloud Run Deployment."""

  class WorkloadType(enum.Enum):
    """The type of the deployment."""

    WORKER_POOL = "worker_pool"
    JOB = "job"
    SERVICE = "service"
    INSTANCE = "instance"

  def __init__(self, args: argparse.Namespace, workload_type: WorkloadType):
    """Initialize the SSH library."""
    self.deployment_name = args.deployment_name
    self.workload_type = workload_type
    self.project = args.project
    self.instance = getattr(args, "instance", None)
    self.container = getattr(args, "container", None)
    self.revision = getattr(args, "revision", None)
    self.region = args.region
    self.release_track = args.release_track
    self.iap_tunnel_url_override = getattr(
        args, "iap_tunnel_url_override", None
    )

    workload_json = self._GetWorkloadJson()
    _ValidateGen2(workload_json)

    if self.workload_type == self.WorkloadType.SERVICE:
      self.revision = self._GetOrValidateRevision(workload_json)
      _ValidateSSHEnabled(workload_json)

    self.service_account = self._GetServiceAccountFromWorkloadJson(
        workload_json
    )
    self.project_number = _GetProjectNumberFromWorkloadJson(workload_json)

    if self._UseCloudRunDomainOverride():
      self.iap_tunnel_url_override = constants.SSH_URL_TEMPLATE.format(
          region=self.region
      )

  def _GetOrValidateRevision(self, workload_json):
    """Returns a valid revision or raises an error.

    Args:
      workload_json: dict, The JSON representation of the Cloud Run workload.

    Returns:
      str, The revision name.

    Raises:
      ValueError: If the revision is invalid, or if there is a traffic split and
        the user does not specify a revision when prompted.
    """
    revisions_serving_traffic = workload_json.get(constants.STATUS, {}).get(
        constants.TRAFFIC, []
    )
    active_revisions = [
        t.get(constants.REVISION_NAME)
        for t in revisions_serving_traffic
        if t.get(constants.PERCENT, 0) > 0
    ]

    if not active_revisions:
      raise ValueError("No serving revisions found for the service.")

    if self.revision and self.revision in active_revisions:
      return self.revision
    if self.revision:
      raise ValueError(
          f"Revision {self.revision} is invalid, or it is not serving"
          " traffic. Please specify one of the following revisions:"
          f" {active_revisions} using the --revision flag."
      )

    # The user did not specify a revision, and there is a traffic split.
    if len(active_revisions) > 1:
      if not console_io.CanPrompt():
        raise ValueError(
            "There is a traffic split. Please specify a revision from the"
            f" following list: {active_revisions} using the --revision flag."
        )

      idx = console_io.PromptChoice(
          active_revisions,
          message=(
              "There is a traffic split. Please specify which revision to SSH"
              " into:\n"
          ),
          cancel_option=True,
      )
      return active_revisions[idx]
    # The user did not specify a revision, and there is only one active
    # revision.
    return active_revisions[0]

  def _UseCloudRunDomainOverride(self):
    """Returns whether to use the Cloud Run domain override."""
    if self.iap_tunnel_url_override:
      return False
    if self.release_track == base.ReleaseTrack.ALPHA and (
        self.workload_type == self.WorkloadType.INSTANCE
    ):
      return False
    return True

  def _UseQualDomain(self) -> bool:
    """Returns whether to use the Qual domain override."""
    if self.iap_tunnel_url_override is None:
      return False
    return (
        "cloud-run-qual" in self.iap_tunnel_url_override
        or "tunnel-staging" in self.iap_tunnel_url_override
        or "tunnel-testing" in self.iap_tunnel_url_override
    )

  def _GetWorkloadJson(self):
    """Retrieves the JSON representation of the Cloud Run workload."""
    command = execution_utils.ArgsForGcloud()

    track_prefix = (
        [self.release_track.prefix]
        if self.release_track and self.release_track.prefix
        else []
    )
    if self.workload_type == self.WorkloadType.SERVICE:
      command.extend(track_prefix + ["run", "services", "describe"])
    elif self.workload_type == self.WorkloadType.WORKER_POOL:
      command.extend(["beta", "run", "worker-pools", "describe"])
    elif self.workload_type == self.WorkloadType.JOB:
      command.extend(track_prefix + ["run", "jobs", "describe"])
    elif self.workload_type == self.WorkloadType.INSTANCE:
      command.extend(["alpha", "run", "instances", "describe"])
    else:
      raise ValueError(f"Unsupported workload type: {self.workload_type}")

    command.extend([
        self.deployment_name,
        "--region",
        self.region,
        "--project",
        self.project,
        "--format",
        "json",
    ])
    try:
      output = subprocess.check_output(command, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
      raise ValueError(
          f"Error describing deployment: {e.stderr.decode('utf-8')}"
      ) from e
    return json.loads(output)

  def _GetServiceAccountFromWorkloadJson(
      self, workload_json: dict[str, Any]
  ) -> str:
    """Retrieves the service account from the workload JSON.

    Args:
      workload_json: A dict, the JSON representation of the Cloud Run workload.

    Returns:
      A str, the service account name.

    Raises:
      ValueError: If the service account is not found in the workload JSON.
    """
    if self.workload_type == self.WorkloadType.INSTANCE:
      service_account = workload_json.get(constants.SPEC, {}).get(
          constants.SERVICE_ACCOUNT_NAME
      )
    else:
      template = workload_json.get(constants.SPEC, {}).get(
          constants.TEMPLATE, {}
      )
      service_account = template.get(constants.SPEC, {}).get(
          constants.SERVICE_ACCOUNT_NAME
      )
    if not service_account:
      raise ValueError("Service account not found for workload.")
    return service_account

  def HostKeyAlias(self):
    """Returns the host key alias for the SSH connection."""
    return constants.SSH_HOST_KEY_ALIAS

  def GetSshCommandComponents(self):
    """Returns the SSH command components."""
    # The PuTTY binary in gcloud does not natively support passing signed
    # certificates via flags. Since OpenSSH is now officially supported on
    # Windows 10 and later, and Cloud Run SSH is a completely new feature,
    # falling back to OpenSSH on Windows is safe.
    if platforms.OperatingSystem.IsWindows():
      env = ssh.Environment(ssh.Suite.OPENSSH, None, is_windows=True)
    else:
      env = ssh.Environment.Current()
    env.RequireSSH()
    keys = ssh.Keys.FromFilename(env=env)
    keys.EnsureKeysExist(overwrite=False)
    user = constants.SSH_ROOT_USER

    # Note: this actually creates the certificate.
    ssh.GetOsloginState(
        None,
        None,
        user,
        keys.GetPublicKey().ToEntry(),
        None,
        self.release_track,
        cloud_run_params={
            "deployment_name": self.deployment_name,
            "project_id": self.project,
            "region": self.region,
            "service_account": self.service_account,
            "workload_type": self.workload_type,
        },
        env=env,
    )
    cert_file = ssh.CertFileFromCloudRunDeployment(
        project=self.project,
        region=self.region,
        deployment=self.deployment_name,
        workload_type=self.workload_type,
    )
    dest_addr = self.HostKeyAlias()
    remote = ssh.Remote(dest_addr, user)

    ca_keys = self._FetchSshCaPublicKeys()
    if ca_keys:
      known_hosts = ssh.KnownHosts.FromDefaultFile()
      known_hosts.AddCertAuthority(
          host_pattern=dest_addr, ca_public_keys=list(ca_keys)
      )
      known_hosts.Write()

    deployment_name = (
        f"{self.deployment_name}/revisions/{self.revision}"
        if self.revision
        else self.deployment_name
    )
    iap_tunnel_args = CreateSshTunnelArgs(
        self.release_track,
        self.project_number,
        self.project,
        deployment_name,
        self.workload_type,
        self.region,
        self.instance,
        self.container,
        self.iap_tunnel_url_override,
    )

    ssh_helper = ssh_utils.BaseSSHCLIHelper()
    ssh_options = ssh_helper.GetConfig(
        host_key_alias=dest_addr,
        strict_host_key_checking="no",
    )

    return SshCommandComponents(
        env=env,
        remote=remote,
        cert_file=cert_file,
        iap_tunnel_args=iap_tunnel_args,
        options=ssh_options,
        identity_file=keys.key_file,
    )

  def Run(self):
    """Run the SSH command."""

    cmd_components = self.GetSshCommandComponents()

    return ssh.SSHCommand(
        remote=cmd_components.remote,
        cert_file=cmd_components.cert_file,
        iap_tunnel_args=cmd_components.iap_tunnel_args,
        options=cmd_components.options,
        identity_file=cmd_components.identity_file,
    ).Run(cmd_components.env)

  def RunScp(
      self,
      sources,
      destination,
      recursive=False,
      compress=False,
      extra_flags=None,
  ):
    """Run the SCP command."""

    cmd_components = self.GetSshCommandComponents()

    return ssh.SCPCommand(
        sources=sources,
        destination=destination,
        recursive=recursive,
        compress=compress,
        cert_file=cmd_components.cert_file,
        iap_tunnel_args=cmd_components.iap_tunnel_args,
        options=cmd_components.options,
        identity_file=cmd_components.identity_file,
        extra_flags=extra_flags,
    ).Run(cmd_components.env)

  def _FetchSshCaPublicKeys(self) -> Sequence[str] | None:
    """Retrieves the CA public keys for the current region from a gstatic URL.

    Returns:
      A Sequence of strings, where each string is a public key, or None if
      the keys could not be fetched.
    """

    template = (
        constants.SSH_CA_PUBLIC_KEY_URL_QUAL_TEMPLATE
        if self._UseQualDomain()
        else constants.SSH_CA_PUBLIC_KEY_URL_TEMPLATE
    )
    endpoint = template.format(
        region=self.region
    )

    try:
      with core_requests.GetSession() as session:
        response = session.get(endpoint, timeout=10)
    except core_requests.requests.exceptions.RequestException:
      log.debug(
          "Failed to fetch SSH CA public keys from %s.",
          endpoint,
          exc_info=True,
      )
      return None
    else:
      if response.status_code != 200:
        log.debug(
            "Failed to fetch SSH CA public keys from %s. Received status"
            " code: %s",
            endpoint,
            response.status_code,
        )
        return None
      return [k.strip() for k in response.text.splitlines() if k.strip()]



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to describe a Device Run session."""

from googlecloudsdk.api_lib import device_run
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.device_run import resource_args


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Describe(base.DescribeCommand):
  """Describe a Device Run session."""

  @staticmethod
  def Args(parser):
    resource_args.AddSessionResourceArg(parser, 'describe')

  def Run(self, args):
    session_ref = args.CONCEPTS.session.Parse()
    client = device_run.SessionsClient(api_version='v1alpha')
    return client.Get(session_ref)


Describe.detailed_help = {
    'DESCRIPTION':
        'Describe a Device Run session.',
    'EXAMPLES':
        """\
To describe a session named `my-session` in location `us-central1`, run:

  $ {command} my-session --location=us-central1
""",
}



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""API client library for Applications."""

from __future__ import annotations

from collections.abc import Sequence
from functools import partial
import time
from typing import Any
import uuid

from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.cloudresourcemanager import projects_api
from googlecloudsdk.api_lib.design_center import utils
from googlecloudsdk.api_lib.iam import util as iam_api
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.iam import iam_util
from googlecloudsdk.command_lib.projects import util as crm_util
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties


def ConstructServiceAccountName(application_id, uuid_value=None):
  """Constructs a unique service account name with a UUID suffix.

  The service account ID must be between 6 and 30 characters.

  Args:
    application_id: str, The ID of the application.
    uuid_value: uuid.UUID, Optional. A UUID value to use instead of generating
      a new one.

  Returns:
    str, The constructed service account name.
  """
  # Service account IDs must be between 6 and 30 characters.
  # We'll use a prefix from the application_id and append a short UUID.
  uuid_str = str(uuid_value) if uuid_value is not None else str(uuid.uuid4())
  uuid_suffix, *_ = uuid_str.split('-')
  max_app_id_len = 30 - len(uuid_suffix) - 1  # Account for '-' separator
  app_id_part = application_id[:max(1, max_app_id_len)].rstrip('-')

  return f'{app_id_part}-{uuid_suffix}'


class ApplicationsClient(object):
  """Client for Applications in the Design Center API."""

  def __init__(self, release_track=base.ReleaseTrack.ALPHA):
    self.client = utils.GetClientInstance(release_track)
    self.messages = utils.GetMessagesModule(release_track)
    self._service = self.client.projects_locations_spaces_applications

  def _RetryIamOperation(
      self,
      operation_fn,
      error_message_fmt: str,
      error_message_args: Sequence[Any],
      max_retries: int = 5,
      delay_seconds: int = 5,
      sleeper=time.sleep,
  ):
    """Retries an IAM operation on HttpError."""
    last_exception = None
    for i in range(max_retries):
      try:
        operation_fn()
        return True
      except apitools_exceptions.HttpError as e:
        formatted_base_message = error_message_fmt % error_message_args
        log.warning(
            '%s (Attempt %d/%d): %r',
            formatted_base_message,
            i + 1,
            max_retries,
            e,
            exc_info=True,
        )
        if i < max_retries - 1:
          sleeper(delay_seconds)
        last_exception = e
    formatted_base_message = error_message_fmt % error_message_args
    log.error(
        '%s (Failed after %d attempts).',
        formatted_base_message,
        max_retries,
    )
    raise exceptions.Error(
        '%s (Failed after %d attempts).' % (formatted_base_message, max_retries)
    ) from last_exception

  def _AddRolesToServiceAccountOnProject(
      self,
      project_ref,
      *,
      service_account_email,
      roles: Sequence[str],
      project_id,
  ):
    """Adds a list of roles to a service account on a specific project.

    Args:
      project_ref: The project resource reference.
      service_account_email: str, The email address of the service account.
      roles: Sequence[str], A sequence of role names to grant.
      project_id: str, The ID of the project.

    Raises:
      exceptions.Error: If adding any role fails after retries.
    """
    for role_to_grant in roles:
      self._RetryIamOperation(
          lambda pr=project_ref, sae=service_account_email, role=role_to_grant: (
              projects_api.AddIamPolicyBinding(
                  pr,
                  f'serviceAccount:{sae}',
                  role,
              )
          ),
          'Failed to add role %s to service account on project %s',
          (role_to_grant, project_id),
      )

  def _AddServiceAccountRoles(
      self,
      project_parameters: Sequence[Any],
      service_account,
  ):
    """Adds required roles to the service account for each project parameter.

    Args:
      project_parameters: A sequence of project parameter objects.
      service_account: str, The email address of the service account.

    Raises:
      exceptions.Error: If adding any role fails.
    """
    for pp in project_parameters:
      current_project_id = pp.projectId
      current_project_ref = crm_util.ParseProject(current_project_id)
      self._AddRolesToServiceAccountOnProject(
          current_project_ref,
          service_account_email=service_account,
          roles=pp.roles,
          project_id=current_project_id,
      )
      log.status.Print(
          'Successfully added required project roles to service account '
          'for project %s.'% current_project_id
      )

  def _CreateAndConfigureServiceAccount(
      self,
      *,
      application_id,
      name,
      project,
      service_account,
      user_account,
      iam_client,
      iam_messages,
  ):
    """Creates and configures a service account for application deployment.

    Args:
      application_id: str, The ID of the application.
      name: str, The full resource name of the Application.
      project: str, The project ID.
      service_account: str, The email address of the service account, or None.
      user_account: str, The email address of the user.
      iam_client: The API client for IAM service accounts.
      iam_messages: The API messages module for IAM service accounts.

    Returns:
      str, The fully qualified service account resource name on success, or None
      on failure.
    Raises:
      exceptions.Error: If service account creation fails for reasons other than
      409.
    """
    # Check if the user has permission to create service accounts.
    project_ref = crm_util.ParseProject(project)
    response = projects_api.TestIamPermissions(
        project_ref, ['iam.serviceAccounts.create']
    )
    # Access the 'permissions' attribute of the response object
    allowed_permissions = (response.permissions or [])
    if 'iam.serviceAccounts.create' not in allowed_permissions:
      raise exceptions.Error(
          'User does not have permission to create service accounts in project'
          f' {project}. Required permission: iam.serviceAccounts.create'
      )

    sa_email = service_account
    if service_account is not None:
      sa_name = service_account.split('@')[0]
    else:
      sa_name = ConstructServiceAccountName(application_id)
      sa_email = f'{sa_name}@{project}.iam.gserviceaccount.com'

    # Describe the application to identify the required IAM roles.
    app_details = self._service.Get(
        self.messages.DesigncenterProjectsLocationsSpacesApplicationsGetRequest(
            name=name
        )
    )
    project_parameters = app_details.projectParameters
    log.status.Print(
        'Successfully described application and retrieved required roles.'
    )

    # Create the service account.
    log.status.Print(
        'Creating service account for application: %s' % application_id
    )
    iam_client.projects_serviceAccounts.Create(
        iam_messages.IamProjectsServiceAccountsCreateRequest(
            name=iam_util.ProjectToProjectResourceName(project),
            createServiceAccountRequest=iam_messages.CreateServiceAccountRequest(
                accountId=sa_name,
                serviceAccount=iam_messages.ServiceAccount(
                    displayName=f'Service account for {sa_email}'
                ),
            ),
        )
    )
    log.status.Print('Successfully created service account: %s' % sa_email)

    self._AddServiceAccountRoles(project_parameters, sa_email)

    def _GrantActAsToPrincipal(sa_email, member):
      """Helper function to grant actAs permission to a given member."""

      sa_resource_name = iam_util.EmailToAccountResourceName(sa_email)

      sa_policy = iam_client.projects_serviceAccounts.GetIamPolicy(
          iam_messages.IamProjectsServiceAccountsGetIamPolicyRequest(
              resource=sa_resource_name
          )
      )

      iam_util.AddBindingToIamPolicy(
          iam_messages.Binding,
          sa_policy,
          member,
          'roles/iam.serviceAccountUser',
      )

      iam_client.projects_serviceAccounts.SetIamPolicy(
          iam_messages.IamProjectsServiceAccountsSetIamPolicyRequest(
              resource=sa_resource_name,
              setIamPolicyRequest=iam_messages.SetIamPolicyRequest(
                  policy=sa_policy
              ),
          )
      )

    def _GrantActAsToUser(sa_email, user_account):
      """Grant `actAs` permission to the user."""
      member = f'user:{user_account}'
      _GrantActAsToPrincipal(sa_email, member)

    def _GrantActAsToAdcAgent(sa_email, adc_service_agent):
      """Grant `actAs` permission to the ADC service agent."""
      member = f'serviceAccount:{adc_service_agent}'
      _GrantActAsToPrincipal(sa_email, member)

    grant_user_func = partial(_GrantActAsToUser, sa_email, user_account)
    self._RetryIamOperation(
        grant_user_func,
        'Failed to grant actAs permission to user %s',
        (sa_email, user_account),
    )
    log.status.Print(
        'Successfully granted actAs permission to user %s.'% user_account
    )

    project_number = projects_api.Get(
        crm_util.ParseProject(project)
    ).projectNumber
    p4sa_host = utils.GetP4saHost()
    adc_service_agent = f'service-{project_number}@{p4sa_host}'

    grant_agent_func = partial(
        _GrantActAsToAdcAgent, sa_email, adc_service_agent
    )
    self._RetryIamOperation(
        grant_agent_func,
        'Failed to grant actAs permission to ADC service agent',
        (sa_email, adc_service_agent),
    )
    log.status.Print(
        'Successfully granted actAs permission to ADC service agent.'
    )
    return f'projects/{project}/serviceAccounts/{sa_email}'

  def ImportIac(self, name, gcs_uri=None, iac_module=None,
                allow_partial_import=False, validate_iac=False):
    """Calls the ImportApplicationIaC RPC.

    Args:
      name: str, The full resource name of the Application.
      gcs_uri: str, The GCS URI of the IaC source.
      iac_module: messages.IaCModule, The IaCModule object.
      allow_partial_import: bool, Whether to allow partial imports.
      validate_iac: bool, Whether to only validate the IaC.

    Returns:
      The response from the API call.
    """
    if not name:
      raise ValueError('Application name cannot be empty or None.')

    import_iac_request = self.messages.ImportApplicationIaCRequest(
        allowPartialImport=allow_partial_import,
        validateIac=validate_iac)

    if gcs_uri:
      import_iac_request.gcsUri = gcs_uri
    elif iac_module:
      import_iac_request.iacModule = iac_module

    request = (
        self.messages.DesigncenterProjectsLocationsSpacesApplicationsImportIaCRequest(
            name=name,
            importApplicationIaCRequest=import_iac_request))

    return self._service.ImportIaC(request)

  def DeployApplication(
      self,
      name: str,
      *,
      replace: bool = False,
      worker_pool: str | None = None,
      service_account: str | None = None,
      create_sa: bool = False,
  ) -> Any | None:
    """Calls the DeployApplication RPC.

    Args:
      name: str, The full resource name of the Application.
      replace: bool, Flag to update the existing deployment.
      worker_pool: str, The user-specified Worker Pool resource.
      service_account: str | None, The email address of the service account.
      create_sa: bool, Whether to create a new service account.

    Returns:
      The response from the API call.
    """
    if not name:
      raise ValueError('Application name cannot be empty or None.')
    application_id = name.split('/')[-1]
    project = properties.VALUES.core.project.Get()
    deploy_service_account = service_account
    if create_sa:
      user_account = properties.VALUES.core.account.Get()
      iam_client, iam_messages = iam_api.GetClientAndMessages()
      configured_sa = self._CreateAndConfigureServiceAccount(
          application_id=application_id,
          name=name,
          project=project,
          service_account=service_account,
          user_account=user_account,
          iam_client=iam_client,
          iam_messages=iam_messages
      )
      deploy_service_account = configured_sa

    return self._service.Deploy(
        self.messages.DesigncenterProjectsLocationsSpacesApplicationsDeployRequest(
            name=name,
            deployApplicationRequest=self.messages.DeployApplicationRequest(
                replace=replace,
                workerPool=worker_pool,
                serviceAccount=deploy_service_account,
            ),
        )
    )

  def GetApplication(self, name: str) -> Any | None:
    """Calls the GetApplication RPC."""
    if not name:
      raise ValueError('Application name cannot be empty or None.')
    return self._service.Get(
        self.messages.DesigncenterProjectsLocationsSpacesApplicationsGetRequest(
            name=name
        )
    )

  def PreviewApplication(
      self,
      name: str,
      worker_pool: str | None = None,
      service_account: str | None = None,
      create_sa: bool = False,
  ):
    """Calls the PreviewApplication RPC.

    Args:
      name: str, The full resource name of the Application.
      worker_pool: str, The user-specified Worker Pool resource.
      service_account: str, The email address of the service account.
      create_sa: bool, Whether to create a new service account.

    Returns:
      The response from the API call.
    """
    if not name:
      raise ValueError('Application name cannot be empty or None.')
    application_id = name.split('/')[-1]
    project = properties.VALUES.core.project.Get()
    preview_service_account = service_account
    if create_sa:
      user_account = properties.VALUES.core.account.Get()
      iam_client, iam_messages = iam_api.GetClientAndMessages()
      configured_sa = self._CreateAndConfigureServiceAccount(
          application_id=application_id,
          name=name,
          project=project,
          service_account=service_account,
          user_account=user_account,
          iam_client=iam_client,
          iam_messages=iam_messages
      )
      preview_service_account = configured_sa

    return self._service.Preview(
        self.messages.DesigncenterProjectsLocationsSpacesApplicationsPreviewRequest(
            name=name,
            previewApplicationRequest=self.messages.PreviewApplicationRequest(
                workerPool=worker_pool,
                serviceAccount=preview_service_account,
            ),
        )
    )

  def GenerateAssessmentReport(self, name: str) -> Any | None:
    """Calls the GenerateAssessmentReport RPC.

    Args:
      name: The full resource name of the Application.

    Returns:
      The response from the API call.
    """
    if not name:
      raise ValueError('Application name cannot be empty or None.')

    return self._service.GenerateAssessmentReport(
        self.messages.DesigncenterProjectsLocationsSpacesApplicationsGenerateAssessmentReportRequest(
            name=name,
            generateApplicationAssessmentReportRequest=self.messages.GenerateApplicationAssessmentReportRequest(),
        )
    )




# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common flags for workforce pools commands."""

from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base


def AddParentFlags(parser, verb):
  parser.add_argument(
      '--organization',
      help='The parent organization of the workforce pool{0} to {1}.'.format(
          's' if verb == 'list' else '', verb
      ),
      required=True,
  )


def AddLocationFlag(parser, verb):
  parser.add_argument(
      '--location',
      help='The location of the workforce pool{0} to {1}.'.format(
          's' if verb == 'list' else '', verb
      ),
      required=True,
  )


def ParseLocation(args):
  if not args.IsSpecified('location'):
    return 'locations/global'
  return 'locations/{}'.format(args.location)


def AddClearableExtendedAttributesOAuth2Client():
  """Creates an ArgumentGroup for clearable OAuth2Client attributes.

  Used for the update-oidc command.

  Returns:
    A list containing a base.ArgumentGroup for clearable OAuth2Client
    attributes.
  """
  clear_extended_attributes_config_arg = base.Argument(
      '--clear-extended-attributes-config',
      dest='clear_extended_attributes_config',
      action='store_true',
      required=False,
      help='Clear the extended attributes configuration.',
  )

  clearable_extended_attributes_group = base.ArgumentGroup(
      mutex=True,
  )
  clearable_extended_attributes_group.AddArgument(
      clear_extended_attributes_config_arg
  )
  clearable_extended_attributes_group.AddArgument(
      ExtendedAttributesOAuth2ClientAttributesGroup(required=False)
  )

  return [clearable_extended_attributes_group]


def AddExtendedAttributesOAuth2Client():
  """Creates an ArgumentGroup for ExtendedAttributesOAuth2Client Attributes.

  Used for the create-oidc command.

  Returns:
    A list containing a base.ArgumentGroup for ExtendedAttributesOAuth2Client
    attributes.
  """
  return [
      ExtendedAttributesOAuth2ClientAttributesGroup(),
  ]


def ExtendedAttributesOAuth2ClientAttributesGroup(required=True):
  """Creates an ArgumentGroup for ExtendedAttributesOAuth2Client Attributes."""
  extended_attributes_client_id_arg = base.Argument(
      '--extended-attributes-client-id',
      dest='extended_attributes_client_id',
      type=str,
      required=required,
      metavar='EXTENDED_ATTRIBUTES_CLIENT_ID',
      help=(
          'The OAuth 2.0 client ID for retrieving extended attributes from the'
          ' identity provider. Required to get extended group memberships for'
          ' a subset of Google Cloud products. The'
          ' --extended-attributes-client-id flag is restricted. We suggest you'
          ' use SCIM instead.'
      ),
  )
  extended_attributes_client_secret_value_arg = base.Argument(
      '--extended-attributes-client-secret-value',
      dest='extended_attributes_client_secret_value',
      type=str,
      required=required,
      metavar='EXTENDED_ATTRIBUTES_CLIENT_SECRET_VALUE',
      help=(
          'The OAuth 2.0 client secret for retrieving extended attributes from'
          ' the identity provider. Required to get extended group memberships'
          ' for a subset of Google Cloud products. The'
          ' --extended-attributes-client-secret-value flag is restricted. We'
          ' suggest you use SCIM instead.'
      ),
  )
  extended_attributes_issuer_uri_arg = base.Argument(
      '--extended-attributes-issuer-uri',
      dest='extended_attributes_issuer_uri',
      type=str,
      required=required,
      metavar='EXTENDED_ATTRIBUTES_ISSUER_URI',
      help=(
          "OIDC identity provider's issuer URI. Must be a valid URI using"
          ' the `https` scheme. Required to get the OIDC discovery'
          ' document. The --extended-attributes-issuer-uri flag is restricted.'
          ' We suggest you use SCIM instead.'
      ),
  )
  # Adding this flag as a ArgList to hide `AZURE_AD_GROUPS_DISPLAY_NAME` from
  # the end user. Currently there is no other way to hide new enum choices.
  # These flags will move back to enum types once feature is ready for launch
  extended_attributes_type_arg = base.Argument(
      '--extended-attributes-type',
      dest='extended_attributes_type',
      type=arg_parsers.ArgList(
          choices=[
              'azure-ad-groups-id',
          ],
          max_length=1,
          min_length=1,
      ),
      required=required,
      metavar='EXTENDED_ATTRIBUTES_TYPE',
      help=(
          'Represents the identity provider and type of claims that should'
          ' be fetched. The --extended-attributes-type flag is restricted. We'
          ' suggest you use SCIM instead.'
      ),
  )
  extended_attributes_filter_arg = base.Argument(
      '--extended-attributes-filter',
      dest='extended_attributes_filter',
      type=str,
      required=False,
      metavar='EXTENDED_ATTRIBUTES_FILTER',
      help=(
          'The filter used to request specific records from the IdP. By'
          ' default, all of the groups that are associated with a user are'
          ' fetched. For Microsoft Entra ID, you can add `$search` query'
          ' parameters using [Keyword Query Language]'
          ' (https://learn.microsoft.com/en-us/sharepoint/dev/general-development/keyword-query-language-kql-syntax-reference).'
          ' To learn more about `$search` querying in Microsoft Entra ID, see'
          ' [Use the `$search` query parameter]'
          ' (https://learn.microsoft.com/en-us/graph/search-query-parameter).'
          ' \n\nAdditionally, Workforce Identity Federation automatically adds'
          ' the following [`$filter` query parameters]'
          ' (https://learn.microsoft.com/en-us/graph/filter-query-parameter),'
          ' based on the value of `attributes_type`. Values passed to `filter`'
          ' are converted to `$search` query parameters. Additional `$filter`'
          ' query parameters cannot be added using this field. \n\n*'
          ' `AZURE_AD_GROUPS_ID`: `securityEnabled` filter is applied. The'
          ' --extended-attributes-filter flag is restricted. We suggest you use'
          ' SCIM instead.'
      ),
  )
  create_extended_attributes_group = base.ArgumentGroup()
  create_extended_attributes_group.AddArgument(
      extended_attributes_client_id_arg
  )
  create_extended_attributes_group.AddArgument(
      extended_attributes_client_secret_value_arg
  )
  create_extended_attributes_group.AddArgument(
      extended_attributes_issuer_uri_arg
  )
  create_extended_attributes_group.AddArgument(extended_attributes_type_arg)
  create_extended_attributes_group.AddArgument(extended_attributes_filter_arg)

  return create_extended_attributes_group



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to wait for Device Run operation completion."""

from googlecloudsdk.api_lib import device_run
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.device_run import resource_args


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Wait(base.SilentCommand):
  """Wait for asynchronous operation to complete."""

  @staticmethod
  def Args(parser):
    resource_args.AddOperationResourceArg(parser, 'wait for')

  def Run(self, args):
    operation_ref = args.CONCEPTS.operation.Parse()
    client = device_run.OperationsClient(api_version='v1alpha')
    poller = device_run.DeviceRunOperationPoller(
        resource_service=None,
        operations_service=client.service,
        resource_ref=None,
    )
    waiter.WaitFor(
        poller,
        operation_ref,
        'Waiting for operation [{}] to complete.'.format(operation_ref.Name()),
    )



#!/usr/bin/env python
"""The BigQuery CLI extract command."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Optional

from absl import flags

import bq_flags
from clients import client_job
from clients import utils as bq_client_utils
from frontend import bigquery_command
from frontend import bq_cached_client
from frontend import flags as frontend_flags
from frontend import utils as frontend_utils
from frontend import utils_flags
from frontend import utils_formatting

# These aren't relevant for user-facing docstrings:
# pylint: disable=g-doc-return-or-yield
# pylint: disable=g-doc-args


class Extract(bigquery_command.BigqueryCmd):
  usage = """extract <source_table> <destination_uris>"""
  command = 'extract'

  def __init__(self, name: str, fv: flags.FlagValues):
    super(Extract, self).__init__(name, fv)
    flags.DEFINE_string(
        'field_delimiter',
        None,
        'The character that indicates the boundary between columns in the '
        'output file. "\\t" and "tab" are accepted names for tab. '
        'Not applicable when extracting models.',
        short_name='F',
        flag_values=fv,
    )
    flags.DEFINE_enum(
        'destination_format',
        None,
        [
            'CSV',
            'NEWLINE_DELIMITED_JSON',
            'AVRO',
            'PARQUET',
            'ML_TF_SAVED_MODEL',
            'ML_XGBOOST_BOOSTER',
        ],
        'The extracted file format. Format CSV, NEWLINE_DELIMITED_JSON, '
        'PARQUET and AVRO are applicable for extracting tables. '
        'Formats ML_TF_SAVED_MODEL and ML_XGBOOST_BOOSTER are applicable for '
        'extracting models. The default value for tables is CSV. Tables with '
        'nested or repeated fields cannot be exported as CSV. The default '
        'value for models is ML_TF_SAVED_MODEL.',
        flag_values=fv,
    )
    flags.DEFINE_integer(
        'trial_id',
        None,
        '1-based ID of the trial to be exported from a hyperparameter tuning '
        'model. The default_trial_id will be exported if not specified. This '
        'does not apply for models not trained with hyperparameter tuning.',
        flag_values=fv,
    )
    flags.DEFINE_boolean(
        'add_serving_default_signature',
        None,
        'Whether to add serving_default signature for export BigQuery ML '
        'trained tf based models.',
        flag_values=fv,
    )
    flags.DEFINE_enum(
        'compression',
        'NONE',
        ['GZIP', 'DEFLATE', 'SNAPPY', 'ZSTD', 'NONE'],
        'The compression type to use for exported files. Possible values '
        'include GZIP, DEFLATE, SNAPPY, ZSTD, and NONE. The default value is '
        'None. Not applicable when extracting models.',
        flag_values=fv,
    )
    flags.DEFINE_boolean(
        'print_header',
        None,
        'Whether to print header rows for formats that '
        'have headers. Prints headers by default.'
        'Not applicable when extracting models.',
        flag_values=fv,
    )
    flags.DEFINE_boolean(
        'use_avro_logical_types',
        None,
        'If destinationFormat is set to "AVRO", this flag indicates whether to '
        'enable extracting applicable column types (such as TIMESTAMP) to '
        'their corresponding AVRO logical types (timestamp-micros), instead of '
        'only using their raw types (avro-long). '
        'Not applicable when extracting models.',
        flag_values=fv,
    )
    flags.DEFINE_boolean(
        'model',
        False,
        'Extract model with this model ID.',
        short_name='m',
        flag_values=fv,
    )
    self.reservation_id_for_a_job_flag = (
        frontend_flags.define_reservation_id_for_a_job(flag_values=fv)
    )
    self._ProcessCommandRc(fv)

  def RunWithArgs(
      self, identifier: str, destination_uris: str
  ) -> Optional[int]:
    """Perform an extract operation of source into destination_uris.

    Usage:
      extract <source_table> <destination_uris>

    Use -m option to extract a source_model.

    Examples:
      bq extract ds.table gs://mybucket/table.csv
      bq extract -m ds.model gs://mybucket/model

    Arguments:
      source_table: Source table to extract.
      source_model: Source model to extract.
      destination_uris: One or more Google Cloud Storage URIs, separated by
        commas.
    """
    client = bq_cached_client.Client.Get()
    kwds = {
        'job_id': utils_flags.get_job_id_from_flags(),
    }
    if bq_flags.LOCATION.value:
      kwds['location'] = bq_flags.LOCATION.value
    if self.reservation_id_for_a_job_flag.present:
      kwds['reservation_id'] = self.reservation_id_for_a_job_flag.value

    if self.m:
      reference = bq_client_utils.GetModelReference(
          id_fallbacks=client, identifier=identifier
      )
    else:
      reference = bq_client_utils.GetTableReference(
          id_fallbacks=client,
          identifier=identifier,
          allow_pcnt_identifier_format=True,
      )
    job = client_job.Extract(
        client,
        reference,
        destination_uris,
        print_header=self.print_header,
        field_delimiter=frontend_utils.NormalizeFieldDelimiter(
            self.field_delimiter
        ),
        destination_format=self.destination_format,
        trial_id=self.trial_id,
        add_serving_default_signature=self.add_serving_default_signature,
        compression=self.compression,
        use_avro_logical_types=self.use_avro_logical_types,
        **kwds,
    )
    if bq_flags.SYNCHRONOUS_MODE.value:
      # If we are here, the job succeeded, but print warnings if any.
      frontend_utils.PrintJobMessages(utils_formatting.format_job_info(job))
    else:
      self.PrintJobStartInfo(job)



"""Generated client library for externalexposure version v1alpha."""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.py import base_api
from googlecloudsdk.generated_clients.apis.externalexposure.v1alpha import externalexposure_v1alpha_messages as messages


class ExternalexposureV1alpha(base_api.BaseApiClient):
  """Generated client library for service externalexposure version v1alpha."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://externalexposure.googleapis.com/'
  MTLS_BASE_URL = 'https://externalexposure.mtls.googleapis.com/'

  _PACKAGE = 'externalexposure'
  _SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v1alpha'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'ExternalexposureV1alpha'
  _URL_VERSION = 'v1alpha'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new externalexposure handle."""
    url = url or self.BASE_URL
    super(ExternalexposureV1alpha, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.folders_locations_operations = self.FoldersLocationsOperationsService(self)
    self.folders_locations = self.FoldersLocationsService(self)
    self.folders = self.FoldersService(self)
    self.organizations_locations_operations = self.OrganizationsLocationsOperationsService(self)
    self.organizations_locations = self.OrganizationsLocationsService(self)
    self.organizations = self.OrganizationsService(self)
    self.projects_locations_operations = self.ProjectsLocationsOperationsService(self)
    self.projects_locations = self.ProjectsLocationsService(self)
    self.projects = self.ProjectsService(self)

  class FoldersLocationsOperationsService(base_api.BaseApiService):
    """Service class for the folders_locations_operations resource."""

    _NAME = 'folders_locations_operations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.FoldersLocationsOperationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Cancel(self, request, global_params=None):
      r"""Starts asynchronous cancellation on a long-running operation. The server makes a best effort to cancel the operation, but success is not guaranteed. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`. Clients can use Operations.GetOperation or other methods to check whether the cancellation succeeded or whether the operation completed despite cancellation. On successful cancellation, the operation is not deleted; instead, it becomes an operation with an Operation.error value with a google.rpc.Status.code of `1`, corresponding to `Code.CANCELLED`.

      Args:
        request: (ExternalexposureFoldersLocationsOperationsCancelRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Cancel')
      return self._RunMethod(
          config, request, global_params=global_params)

    Cancel.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations/{locationsId}/operations/{operationsId}:cancel',
        http_method='POST',
        method_id='externalexposure.folders.locations.operations.cancel',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}:cancel',
        request_field='cancelOperationRequest',
        request_type_name='ExternalexposureFoldersLocationsOperationsCancelRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a long-running operation. This method indicates that the client is no longer interested in the operation result. It does not cancel the operation. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`.

      Args:
        request: (ExternalexposureFoldersLocationsOperationsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations/{locationsId}/operations/{operationsId}',
        http_method='DELETE',
        method_id='externalexposure.folders.locations.operations.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureFoldersLocationsOperationsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the latest state of a long-running operation. Clients can use this method to poll the operation result at intervals as recommended by the API service.

      Args:
        request: (ExternalexposureFoldersLocationsOperationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations/{locationsId}/operations/{operationsId}',
        http_method='GET',
        method_id='externalexposure.folders.locations.operations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureFoldersLocationsOperationsGetRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists operations that match the specified filter in the request. If the server doesn't support this method, it returns `UNIMPLEMENTED`.

      Args:
        request: (ExternalexposureFoldersLocationsOperationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListOperationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations/{locationsId}/operations',
        http_method='GET',
        method_id='externalexposure.folders.locations.operations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken', 'returnPartialSuccess'],
        relative_path='v1alpha/{+name}/operations',
        request_field='',
        request_type_name='ExternalexposureFoldersLocationsOperationsListRequest',
        response_type_name='ListOperationsResponse',
        supports_download=False,
    )

  class FoldersLocationsService(base_api.BaseApiService):
    """Service class for the folders_locations resource."""

    _NAME = 'folders_locations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.FoldersLocationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Gets information about a location.

      Args:
        request: (ExternalexposureFoldersLocationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Location) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations/{locationsId}',
        http_method='GET',
        method_id='externalexposure.folders.locations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureFoldersLocationsGetRequest',
        response_type_name='Location',
        supports_download=False,
    )

    def GetScanMetrics(self, request, global_params=None):
      r"""Gets the latest scan metrics for a given scope. When called for a folder or organization, the metrics returned are an aggregation of metrics from all projects within the scope. This is a singleton resource.

      Args:
        request: (ExternalexposureFoldersLocationsGetScanMetricsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ScanMetrics) The response message.
      """
      config = self.GetMethodConfig('GetScanMetrics')
      return self._RunMethod(
          config, request, global_params=global_params)

    GetScanMetrics.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations/{locationsId}/scanMetrics',
        http_method='GET',
        method_id='externalexposure.folders.locations.getScanMetrics',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureFoldersLocationsGetScanMetricsRequest',
        response_type_name='ScanMetrics',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists information about the supported locations for this service. This method lists locations based on the resource scope provided in the ListLocationsRequest.name field: * **Global locations**: If `name` is empty, the method lists the public locations available to all projects. * **Project-specific locations**: If `name` follows the format `projects/{project}`, the method lists locations visible to that specific project. This includes public, private, or other project-specific locations enabled for the project. For gRPC and client library implementations, the resource name is passed as the `name` field. For direct service calls, the resource name is incorporated into the request path based on the specific service implementation and version.

      Args:
        request: (ExternalexposureFoldersLocationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListLocationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/folders/{foldersId}/locations',
        http_method='GET',
        method_id='externalexposure.folders.locations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['extraLocationTypes', 'filter', 'pageSize', 'pageToken'],
        relative_path='v1alpha/{+name}/locations',
        request_field='',
        request_type_name='ExternalexposureFoldersLocationsListRequest',
        response_type_name='ListLocationsResponse',
        supports_download=False,
    )

  class FoldersService(base_api.BaseApiService):
    """Service class for the folders resource."""

    _NAME = 'folders'

    def __init__(self, client):
      super(ExternalexposureV1alpha.FoldersService, self).__init__(client)
      self._upload_configs = {
          }

  class OrganizationsLocationsOperationsService(base_api.BaseApiService):
    """Service class for the organizations_locations_operations resource."""

    _NAME = 'organizations_locations_operations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.OrganizationsLocationsOperationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Cancel(self, request, global_params=None):
      r"""Starts asynchronous cancellation on a long-running operation. The server makes a best effort to cancel the operation, but success is not guaranteed. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`. Clients can use Operations.GetOperation or other methods to check whether the cancellation succeeded or whether the operation completed despite cancellation. On successful cancellation, the operation is not deleted; instead, it becomes an operation with an Operation.error value with a google.rpc.Status.code of `1`, corresponding to `Code.CANCELLED`.

      Args:
        request: (ExternalexposureOrganizationsLocationsOperationsCancelRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Cancel')
      return self._RunMethod(
          config, request, global_params=global_params)

    Cancel.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations/{locationsId}/operations/{operationsId}:cancel',
        http_method='POST',
        method_id='externalexposure.organizations.locations.operations.cancel',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}:cancel',
        request_field='cancelOperationRequest',
        request_type_name='ExternalexposureOrganizationsLocationsOperationsCancelRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a long-running operation. This method indicates that the client is no longer interested in the operation result. It does not cancel the operation. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`.

      Args:
        request: (ExternalexposureOrganizationsLocationsOperationsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='DELETE',
        method_id='externalexposure.organizations.locations.operations.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureOrganizationsLocationsOperationsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the latest state of a long-running operation. Clients can use this method to poll the operation result at intervals as recommended by the API service.

      Args:
        request: (ExternalexposureOrganizationsLocationsOperationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='GET',
        method_id='externalexposure.organizations.locations.operations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureOrganizationsLocationsOperationsGetRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists operations that match the specified filter in the request. If the server doesn't support this method, it returns `UNIMPLEMENTED`.

      Args:
        request: (ExternalexposureOrganizationsLocationsOperationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListOperationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations/{locationsId}/operations',
        http_method='GET',
        method_id='externalexposure.organizations.locations.operations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken', 'returnPartialSuccess'],
        relative_path='v1alpha/{+name}/operations',
        request_field='',
        request_type_name='ExternalexposureOrganizationsLocationsOperationsListRequest',
        response_type_name='ListOperationsResponse',
        supports_download=False,
    )

  class OrganizationsLocationsService(base_api.BaseApiService):
    """Service class for the organizations_locations resource."""

    _NAME = 'organizations_locations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.OrganizationsLocationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Gets information about a location.

      Args:
        request: (ExternalexposureOrganizationsLocationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Location) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations/{locationsId}',
        http_method='GET',
        method_id='externalexposure.organizations.locations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureOrganizationsLocationsGetRequest',
        response_type_name='Location',
        supports_download=False,
    )

    def GetScanMetrics(self, request, global_params=None):
      r"""Gets the latest scan metrics for a given scope. When called for a folder or organization, the metrics returned are an aggregation of metrics from all projects within the scope. This is a singleton resource.

      Args:
        request: (ExternalexposureOrganizationsLocationsGetScanMetricsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ScanMetrics) The response message.
      """
      config = self.GetMethodConfig('GetScanMetrics')
      return self._RunMethod(
          config, request, global_params=global_params)

    GetScanMetrics.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations/{locationsId}/scanMetrics',
        http_method='GET',
        method_id='externalexposure.organizations.locations.getScanMetrics',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureOrganizationsLocationsGetScanMetricsRequest',
        response_type_name='ScanMetrics',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists information about the supported locations for this service. This method lists locations based on the resource scope provided in the ListLocationsRequest.name field: * **Global locations**: If `name` is empty, the method lists the public locations available to all projects. * **Project-specific locations**: If `name` follows the format `projects/{project}`, the method lists locations visible to that specific project. This includes public, private, or other project-specific locations enabled for the project. For gRPC and client library implementations, the resource name is passed as the `name` field. For direct service calls, the resource name is incorporated into the request path based on the specific service implementation and version.

      Args:
        request: (ExternalexposureOrganizationsLocationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListLocationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/organizations/{organizationsId}/locations',
        http_method='GET',
        method_id='externalexposure.organizations.locations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['extraLocationTypes', 'filter', 'pageSize', 'pageToken'],
        relative_path='v1alpha/{+name}/locations',
        request_field='',
        request_type_name='ExternalexposureOrganizationsLocationsListRequest',
        response_type_name='ListLocationsResponse',
        supports_download=False,
    )

  class OrganizationsService(base_api.BaseApiService):
    """Service class for the organizations resource."""

    _NAME = 'organizations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.OrganizationsService, self).__init__(client)
      self._upload_configs = {
          }

  class ProjectsLocationsOperationsService(base_api.BaseApiService):
    """Service class for the projects_locations_operations resource."""

    _NAME = 'projects_locations_operations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.ProjectsLocationsOperationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Cancel(self, request, global_params=None):
      r"""Starts asynchronous cancellation on a long-running operation. The server makes a best effort to cancel the operation, but success is not guaranteed. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`. Clients can use Operations.GetOperation or other methods to check whether the cancellation succeeded or whether the operation completed despite cancellation. On successful cancellation, the operation is not deleted; instead, it becomes an operation with an Operation.error value with a google.rpc.Status.code of `1`, corresponding to `Code.CANCELLED`.

      Args:
        request: (ExternalexposureProjectsLocationsOperationsCancelRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Cancel')
      return self._RunMethod(
          config, request, global_params=global_params)

    Cancel.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}:cancel',
        http_method='POST',
        method_id='externalexposure.projects.locations.operations.cancel',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}:cancel',
        request_field='cancelOperationRequest',
        request_type_name='ExternalexposureProjectsLocationsOperationsCancelRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a long-running operation. This method indicates that the client is no longer interested in the operation result. It does not cancel the operation. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`.

      Args:
        request: (ExternalexposureProjectsLocationsOperationsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='DELETE',
        method_id='externalexposure.projects.locations.operations.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureProjectsLocationsOperationsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the latest state of a long-running operation. Clients can use this method to poll the operation result at intervals as recommended by the API service.

      Args:
        request: (ExternalexposureProjectsLocationsOperationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='GET',
        method_id='externalexposure.projects.locations.operations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureProjectsLocationsOperationsGetRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists operations that match the specified filter in the request. If the server doesn't support this method, it returns `UNIMPLEMENTED`.

      Args:
        request: (ExternalexposureProjectsLocationsOperationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListOperationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations/{locationsId}/operations',
        http_method='GET',
        method_id='externalexposure.projects.locations.operations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken', 'returnPartialSuccess'],
        relative_path='v1alpha/{+name}/operations',
        request_field='',
        request_type_name='ExternalexposureProjectsLocationsOperationsListRequest',
        response_type_name='ListOperationsResponse',
        supports_download=False,
    )

  class ProjectsLocationsService(base_api.BaseApiService):
    """Service class for the projects_locations resource."""

    _NAME = 'projects_locations'

    def __init__(self, client):
      super(ExternalexposureV1alpha.ProjectsLocationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Gets information about a location.

      Args:
        request: (ExternalexposureProjectsLocationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Location) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations/{locationsId}',
        http_method='GET',
        method_id='externalexposure.projects.locations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureProjectsLocationsGetRequest',
        response_type_name='Location',
        supports_download=False,
    )

    def GetScanMetrics(self, request, global_params=None):
      r"""Gets the latest scan metrics for a given scope. When called for a folder or organization, the metrics returned are an aggregation of metrics from all projects within the scope. This is a singleton resource.

      Args:
        request: (ExternalexposureProjectsLocationsGetScanMetricsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ScanMetrics) The response message.
      """
      config = self.GetMethodConfig('GetScanMetrics')
      return self._RunMethod(
          config, request, global_params=global_params)

    GetScanMetrics.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations/{locationsId}/scanMetrics',
        http_method='GET',
        method_id='externalexposure.projects.locations.getScanMetrics',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha/{+name}',
        request_field='',
        request_type_name='ExternalexposureProjectsLocationsGetScanMetricsRequest',
        response_type_name='ScanMetrics',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists information about the supported locations for this service. This method lists locations based on the resource scope provided in the ListLocationsRequest.name field: * **Global locations**: If `name` is empty, the method lists the public locations available to all projects. * **Project-specific locations**: If `name` follows the format `projects/{project}`, the method lists locations visible to that specific project. This includes public, private, or other project-specific locations enabled for the project. For gRPC and client library implementations, the resource name is passed as the `name` field. For direct service calls, the resource name is incorporated into the request path based on the specific service implementation and version.

      Args:
        request: (ExternalexposureProjectsLocationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListLocationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha/projects/{projectsId}/locations',
        http_method='GET',
        method_id='externalexposure.projects.locations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['extraLocationTypes', 'filter', 'pageSize', 'pageToken'],
        relative_path='v1alpha/{+name}/locations',
        request_field='',
        request_type_name='ExternalexposureProjectsLocationsListRequest',
        response_type_name='ListLocationsResponse',
        supports_download=False,
    )

  class ProjectsService(base_api.BaseApiService):
    """Service class for the projects resource."""

    _NAME = 'projects'

    def __init__(self, client):
      super(ExternalexposureV1alpha.ProjectsService, self).__init__(client)
      self._upload_configs = {
          }
